classdef VEMalgo < handle
    %==========================================================================
    % Variational expectation-maximization algorithm.
    % 
    %
    % S. Leglaive, R. Badeau, G. Richard. "Student's t source and mixing models 
    % for multichannel audio source separation", submitted for publication in 
    % a journal, 2017. 
    %
    %==========================================================================
    % Copyright (c) 2017, Simon Leglaive, Telecom ParisTech, All Rights Reserved.
    %
    % This code uses the BSD License:
    %
    % Redistribution and use in source and binary forms, with or without
    % modification, are permitted provided that the following conditions are met:
    %
    % * Redistributions of source code must retain the above copyright
    % notice, this list of conditions and the following disclaimer.
    % * Redistributions in binary form must reproduce the above copyright
    % notice, this list of conditions and the following disclaimer in the
    % documentation and/or other materials provided with the distribution.
    % * Neither the name of Telecom ParisTech nor the names of its contributors
    % may be used to endorse or promote products derived from this software without
    % specific prior written permission.
    %
    % THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
    % EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    % WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    % DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
    % DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    % (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    % LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
    % ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    % (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    % SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    %==========================================================================
    
    properties
        
        x; % time domain observations [T x I]
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % PRIOR MODEL PARAMETERS
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Source model parameters:
        lambda2; % Squared Student's t scale parameters {J x 1}[Fj x Nj]
        W; % {J x 1} cell array containing the matrices of NMF spectral templates for each source
        H; % {J x 1} cell array containing the matrices of NMF activations for each source
        alpha_v; % Student's t shape parameter
        
        % Mixing filters model parameters:
        q2; % Squared Student's t scale parameters [La x 1]
        alpha_u; % Student's t shape parameter [La x 1]
        
        % Noise parameters:
        sigma2; % noise variance [I x 1]
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % VARIATIONAL DISTRIBUTION PARAMETERS
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % S:
        s_hat; % s_jfn mean {J x 1}[Fj x Nj]
        gamma; % s_jfn variance {J x 1}[Fj x Nj]
        
        % V:
        nu_v; % v_jfn shape parameter 
        beta; % v_jfn scale parameters {J x 1}[Fj x Nj]       
        
        % A:
        a_hat; % a_ij(t) mean [I x J x La]
        rho % a_ij(t) variance [I x J x La]
        
        % U:
        nu_u; % u_ij(t) shape parameter [La x 1]
        d; % % u_ij(t) scale parameters [I x J x La]
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % OTHERS
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        wlen; % MDCT synthesis window length [J x 1]
        expectError; % expectation of the L2 norm of the difference between the model and the observations [I x 1]
        varFreeEnergy; % objective function
        psiCorr; % troncated autocorrelation of time-frequency MDCT atoms {J x 1}[Fj x La]
        G; % matrix containing the convolution of the MDCT atoms with the posterior mean of the 
        % mixing filters {J x 1}[Fj x (La + wlen - 1) x I]
        s; % [Ls x J] time-domain sources
        sim; % [T x J x I] time-domain source images
    end
    
    methods
        
        function this = VEMalgo(varargin)
            this.Set(varargin{:});
        end
        
        function Set(this,varargin)
            st_input = struct(varargin{:});

            if isfield(st_input, 'x')
                this.x = st_input.x;
            end
            
            if isfield(st_input, 'alpha_u')
                this.alpha_u = st_input.alpha_u;
            end
            
            if isfield(st_input, 'q2')
                this.q2 = st_input.q2;
            end
            
            if isfield(st_input, 'sigma2')
                this.sigma2 = st_input.sigma2;
            end
            
            if isfield(st_input, 'alpha_v')
                this.alpha_v = st_input.alpha_v;
            end
            
            if isfield(st_input, 'W')
                this.W = st_input.W;
            end
            
            if isfield(st_input, 'H')
                this.H = st_input.H;
            end
            
            if isfield(st_input, 'wlen')
                this.wlen = st_input.wlen;
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Definition of some parameters
            
            if isfield(st_input, 'La')
                La = st_input.La;
            end
            
            if isfield(st_input, 'lambda2')
                J = length(st_input.lambda2');
            elseif isfield(st_input, 'W')
                J = numel(st_input.W);
            end
            hop = this.wlen/2;
            [T, I] = size(this.x);
            Ls = T - La + 1;
            F = this.wlen/2;
            N = round(Ls./hop) - 1;
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Initializations
            
            this.lambda2 = cell(J,1);
            for j=1:J
                this.lambda2{j} = ones(F(j), N(j));
            end

            if isfield(st_input, 'W') && isfield(st_input, 'H')
                for j=1:J
                    this.lambda2{j} = this.W{j}*this.H{j};
                end
            end
            
            this.s = zeros(Ls, J);
            this.sim = zeros(T, J, I);
            
            fprintf('...entering InitializePosterior\n');
            this.InitializePosterior(F, N, J, I, La);
            
            fprintf('...entering ComputePsiCorr\n');
            this.ComputePsiCorr;
            
            % Compute mean and variance of g_{ij,fn}(t)
            this.G = cell(J,1);
            for j=1:J
                this.G{j} = zeros(F(j), La + this.wlen(j) - 1, I);
            end
            
            fprintf('...entering ComputeGMat\n');
            this.ComputeGMat;
            
            % Compute expected error
            this.expectError = zeros(I,1);
            this.ComputeExpectError;
                        
        end
       
        function [] = InitializePosterior(this, F, N, J, I, La)

            % V:
            this.nu_v = (this.alpha_v + 1)/2;
            this.beta = cell(J,1);
            for j=1:J
                this.beta{j} = this.alpha_v/2*ones(F(j),N(j));
            end
            
            % S:
            this.gamma = cell(J,1);
            this.s_hat = cell(J,1);
            for j=1:J
                rng(0);
                this.gamma{j} = 1./gamrnd(this.nu_v, 1./this.beta{j}) .* this.lambda2{j}; 
                rng(0);
                this.s_hat{j} = sqrt(this.gamma{j}).*randn(F(j),N(j));
            end
            
            % U:
            this.nu_u = (this.alpha_u + 1)/2;
            this.d = permute(repmat(this.alpha_u/2, 1, I, J), [2,3,1]);
            
            % A:
            this.rho = ones(I,J,La);
            this.a_hat = ones(I,J,La);
            rngCpt = 0;
            for i=1:I
                for j=1:J
                                        
                    rng(rngCpt);
                    u_ij = 1./gamrnd(this.nu_u, 1./permute(this.d(i,j,:), [3,1,2]) ); 
                    
                    this.rho(i,j,:) = u_ij.*this.q2;
                    
                    rng(rngCpt);
                    this.a_hat(i,j,:) = sqrt(permute(this.rho(i,j,:), [3,4,1,2])).*randn(La,1);

                    rngCpt = rngCpt + 1;
                end
            end
            
            % Initialize source and image signals
            this.ImdctMultiProp(1:J);
            this.SrcImageProp(1:J,(1:I));

        end
        
        function [] = ComputePsiCorr(this)
            
            J = numel(this.s_hat);            
            F = cellfun('size', this.s_hat, 1);            
            [T, ~] = size(this.x);
            La = size(this.a_hat,3);
            Ls = T - La + 1;
            
            this.psiCorr = cell(J,1); % {J x 1}[Fj x La]

            for j=1:J
                
                if j>1 && sum(this.wlen(j)==this.wlen(1:j-1))
                    this.psiCorr{j} = this.psiCorr{find(this.wlen(j)==this.wlen(1:j-1),1)};
                else
                    psiCorrTmp = zeros(F(j),La);
                    for f=1:F(j)
                        psi_jfn = this.GetMDCTAtom(f-1,0,Ls,this.wlen(j));
                        tmp = xcorr(psi_jfn, La-1);
                        psiCorrTmp(f,:) = tmp(La:end).';
                    end
                    this.psiCorr{j} = psiCorrTmp;
                end
            end            
            
        end
        
        
        function [] = E_V_Step(this)
            
            J = numel(this.s_hat);
            
            this.nu_v = (this.alpha_v + 1)/2;
            
            for j=1:J
                this.beta{j} = ( (this.s_hat{j}.^2 + this.gamma{j})./(this.lambda2{j} + eps) + this.alpha_v )/2;
            end
            
        end
        
        function [] = E_S_Step(this)
                        
            [~, I] = size(this.x);
            J = numel(this.s_hat);            
            N = cellfun('size', this.s_hat, 2);
            La = size(this.a_hat,3);
            hop = this.wlen/2;
            
            inv_v_post = cellfun(@(x) this.nu_v./x, this.beta, 'uniformoutput', false);
            
            % compute gamma
            for j=1:J
                % sum_t g_{ij,fn}(t)^2
                tmp = permute(sum(this.G{j}.^2,2), [1,3,2]); % F x I
                % sum_t g_{ij,fn}(t)^2 + sum_tau rho_ij(tau)
                tmp = bsxfun(@plus, tmp, permute(sum(this.rho(:,j,:),3), [2,1,3])); % F x I
                % sum_i 1/sigma_i^2 ( sum_t g_{ij,fn}(t)^2 + sum_tau rho_ij(tau) )
                tmp = sum(bsxfun(@times, tmp, permute(1./this.sigma2, [2,1])),2); % F x 1
                
                this.gamma{j} = 1./bsxfun(@plus, inv_v_post{j}./this.lambda2{j}, tmp); % F x N 
            end
            clear tmp

            % compute gradient
            epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
            inv_sigma2_sum_rho = sum(bsxfun(@times, 1./this.sigma2, sum(this.rho,3)),1); % 1 x J
            grad = cell(J,1);
            
            frameInd = cell(J,1);
            
            for j=1:J
                frameInd{j} = zeros(this.wlen(j)+La-1,N(j)); % column n of frameInd contains the indices of the signal to be considered for frame n
                for n=1:N(j)
                    frameInd{j}(:,n) = (n-1)*hop(j) + (1:this.wlen(j)+La-1)';
                end
                epsilonMat = zeros(this.wlen(j)+La-1,N(j),I);
                for i=1:I
                    epsilon_i = epsilon(:,i)/this.sigma2(i); % (wlen + La - 1) x N
                    epsilonMat(:,:,i) = epsilon_i(frameInd{j}); % (wlen + La - 1) x N x I
                end

                grad{j} = this.s_hat{j} .* ( inv_v_post{j}./(this.lambda2{j} + eps) + inv_sigma2_sum_rho(j)) - ...
                    sum(mtimesx(this.G{j}, epsilonMat),3);
            end
            
            % compute descent direction
            w = cell(J,1);
            for j=1:J
                w{j} = this.gamma{j}.*grad{j};
            end
            
            % allocate
            kappa = cell(J,1);
            gradP = cell(J,1);
            
            niter = 10;
            for iter = 1:niter
                
                % compute kappa
                w_mdct = this.ImdctMulti(w); % Ls x J
                w_filt = this.SrcImage(w_mdct); % T x J x I
                w_filt = squeeze(sum(w_filt, 2)); % T x I

                for j=1:J
                    
                    w_filtMat = zeros(this.wlen(j)+La-1,N(j),I); % (wlen + La - 1) x N x I
                    for i=1:I
                        w_filt_i = w_filt(:,i)/this.sigma2(i); 
                        w_filtMat(:,:,i) = w_filt_i(frameInd{j});
                    end
                    
                    kappa{j} = w{j}.* ( inv_v_post{j}./(this.lambda2{j} + eps) + inv_sigma2_sum_rho(j)) + ...
                        sum(mtimesx(this.G{j},w_filtMat),3);
                    
                end
                
                % compute mu
                mu = real((this.cell2vec(w)'*this.cell2vec(grad))/(this.cell2vec(w)'*this.cell2vec(kappa)));
                if mu<0
                    mu = eps;
                end
                
                % update parameters to be estimated
                this.s_hat = cellfun(@(x,y) x - mu*y, this.s_hat, w, 'UniformOutput', false);
                
                % update source images
                this.ImdctMultiProp(1:J);
                this.SrcImageProp(1:J,(1:I));
                
                if iter == niter
                    break;
                end
                
                % compute gradient
                epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
                for j=1:J
                    
                    epsilonMat = zeros(this.wlen(j)+La-1,N(j),I);
                    for i=1:I
                        epsilon_i = epsilon(:,i)/this.sigma2(i); % (wlen + La - 1) x N
                        epsilonMat(:,:,i) = epsilon_i(frameInd{j}); % (wlen + La - 1) x N x I
                    end
                    
                    grad{j} = this.s_hat{j} .* ( inv_v_post{j}./(this.lambda2{j} + eps) + inv_sigma2_sum_rho(j)) - ...
                        sum(mtimesx(this.G{j}, epsilonMat),3);
                end

                % preconditionning
                for j=1:J
                    gradP{j} = this.gamma{j}.*grad{j};
                end

                % compute alpha
                alpha_pcg = - (this.cell2vec(kappa)'*this.cell2vec(gradP)) / (this.cell2vec(w)'*this.cell2vec(kappa));
                if alpha_pcg<0
                    alpha_pcg = eps;
                end
                
                % compute w
                w = cellfun(@(x,y) x + alpha_pcg*y, gradP, w, 'UniformOutput', false);
                
            end
        end
        
        function [] = E_U_Step(this)
            
            [~, I] = size(this.x);
            J = numel(this.s_hat);            
            
            this.nu_u = (this.alpha_u + 1)/2;

            for i=1:I
                for j=1:J                    
                    this.d(i,j,:) = ( this.alpha_u + ( permute(this.a_hat(i,j,:), [3 1 2]).^2 + ...
                            permute(this.rho(i,j,:), [3 1 2]) )./this.q2 )/2;
                end
            end
            
        end
                        
        function [] = E_A_Step(this)
                        
            [~, I] = size(this.x);
            J = numel(this.s_hat);            
            N = cellfun('size', this.s_hat, 2);
            La = size(this.a_hat,3);
            
            inv_u_post = bsxfun(@times, 1./this.d, permute(this.nu_u, [2,3,1])); % I x J x La
            
            % Compute sum_{f,n} gamma_fn r^{psi,psi}_fn
            gammaTT = zeros(La, J);
            for j=1:J
                gamma_jfn = this.gamma{j};
                for t=1:La
                    psiCorr_t = this.psiCorr{j}(:,t);
                    psiCorr_t = repmat(psiCorr_t, [1, N(j)]);
                    gammaTT(t,j) = psiCorr_t(:).'*gamma_jfn(:);
                end
            end
            clear psiCorr_t;
            
            % compute r_ss
            srcCorr = zeros(La,J);
            for j=1:J
                tmp = xcorr(this.s(:,j), La-1);
                srcCorr(:,j) = tmp(La:end);
            end
            
            % compute rho
            tmp1 = bsxfun(@times, 1./this.sigma2, gammaTT(1,:) + srcCorr(1,:)); % I x J
            tmp2 = bsxfun(@times, inv_u_post, permute(1./this.q2,[2,3,1])); % I x J x La
            this.rho = 1./bsxfun(@plus, tmp2, tmp1);
            clear tmp1 tmp2;
            
            Gamma_a_diagPart = bsxfun(@times, 1./this.q2, permute(inv_u_post, [3,1,2])); % La x I x J
            Gamma_a_toepPart = bsxfun(@times, permute(1./this.sigma2,[3,1,2]), permute(srcCorr + gammaTT, [1,3,2])); % La x I x J
            
            this.a_hat = permute(this.a_hat, [3,1,2]);
            this.rho = permute(this.rho, [3,1,2]);
            
            niter = 10;
            for j=1:J
                for i=1:I
                   
                    % compute the difference between x_i and sum_{j' ~= j} s_{ij'}^{im}
                    epsilon_ij = this.x(:,i) - sum(this.sim(:,[1:j-1, j+1:J],i),2); % T x 1
                    
                    % compute cross-correlation s_j and epsilon_ij
                    tmp = xcorr(epsilon_ij, this.s(:,j), La-1);
                    srcEpsilonCorr_ij = tmp(La:end); % La x 1
                    
                    % compute gradient
                    grad_ij = Gamma_a_diagPart(:,i,j).*this.a_hat(:,i,j) + ...
                        toeplitzmult(Gamma_a_toepPart(:,i,j),Gamma_a_toepPart(:,i,j),this.a_hat(:,i,j)) - ...
                        1./this.sigma2(i)*srcEpsilonCorr_ij;
                    
                    % compute descent direction
                    w_ij = this.rho(:,i,j).*grad_ij;
                    
                    for iter = 1:niter
                        
                        kappa_ij = Gamma_a_diagPart(:,i,j).*w_ij + ...
                            toeplitzmult(Gamma_a_toepPart(:,i,j), Gamma_a_toepPart(:,i,j), w_ij);
                        
                        % compute mu
                        mu = real((w_ij'*grad_ij)/(w_ij'*kappa_ij));
                        if mu<0
                            mu = eps;
                        end
                        
                        % update parameters to be estimated
                        this.a_hat(:,i,j) = this.a_hat(:,i,j) - mu*w_ij;
                        
                        if iter == niter
                            break;
                        end
                        
                        % compute gradient
                        grad_ij = Gamma_a_diagPart(:,i,j).*this.a_hat(:,i,j) + ...
                            toeplitzmult(Gamma_a_toepPart(:,i,j),Gamma_a_toepPart(:,i,j),this.a_hat(:,i,j)) - ...
                            1./this.sigma2(i)*srcEpsilonCorr_ij;
                        
                        % preconditionning
                        gradP_ij = this.rho(:,i,j).*grad_ij;
                        
                        % compute alpha
                        alpha_pcg = -real((kappa_ij'*gradP_ij)/(w_ij'*kappa_ij));
                        if alpha_pcg<0
                            alpha_pcg = eps;
                        end
                        
                        % compute w
                        w_ij = gradP_ij + alpha_pcg*w_ij;
                        
                    end
                end
                
                this.a_hat = permute(this.a_hat, [2,3,1]);
                this.SrcImageProp(j,(1:I));
                this.a_hat = permute(this.a_hat, [3,1,2]);
            end
            
            this.a_hat = permute(this.a_hat, [2,3,1]);
            this.rho = permute(this.rho, [2,3,1]);
            this.ComputeGMat;
            
        end 

        function [] = UpdateNoiseVar(this)
            
            [T,~] = size(this.x);
            this.sigma2 = 1/T*this.expectError;
            
        end
        
        function [] = UpdateNoiseVarAnneal(this, iter, niter, begVal, endVal)
            % Due to this annealing procedure, it is possible that the variational free energy decreases.
            % For developing purposes, use the function 'UpdateNoiseVar'
            % which leads to a monotonic increase of the variational free
            % energy.
            
            [~,I] = size(this.x);
            sig = ((sqrt(begVal) * (niter - iter) + sqrt(endVal) * iter) / niter).^2;
            this.sigma2 = sig*ones(I,1);
            
        end
        

        function [] = UpdateLambda(this, b_updW)         
            
            if nargin < 2
                b_updW = 1;
            end

            J = numel(this.s_hat); 
            
            for j=1:J
                this.UpdateNMF(j, 20, b_updW);
                this.lambda2{j} = this.W{j}*this.H{j};
            end
            
        end
                
        function E = UpdateNMF(this, j, iter_num, b_updW)
            
            if nargin < 4 || isempty(b_updW)
                b_updW = 1;
            end
            
            if nargin < 3 || isempty(iter_num)
                iter_num = 10;
            end
            
            V = (this.s_hat{j}.^2 + this.gamma{j})./(this.beta{j}./this.nu_v);
                        
            WH = this.W{j}*this.H{j} + eps;
            
            matOfOnes = ones(size(V));
            
            E = zeros(1, iter_num);
            
            for iter = 1:iter_num
                
                if b_updW
                    
                    % update of W
                    this.W{j} = this.W{j} .* (( V./((WH).^2) )*this.H{j}' + eps) ./ ((1./(WH))*this.H{j}' + eps);
                    
                    % normalization
                    w = sum(this.W{j},1);
                    dd = diag(1./w);
                    this.W{j} = this.W{j}*dd; % Normalization of columns of W
                    dd = diag(w);
                    this.H{j} = dd*this.H{j}; % Energy transfer to H
                    
                    WH = this.W{j}*this.H{j} + eps;
                    
                end

                % update of H
                this.H{j} = this.H{j} .* (this.W{j}'*(V./((WH).^2)) + eps) ./ (this.W{j}'*(1./(WH)) + eps);
                
                WH = this.W{j}*this.H{j} + eps;
                
                % IS divergence:
                E(iter) = sum(sum(V./(WH + eps) - log(V./(WH + eps) + eps) - matOfOnes)); % normalized error

            end
        end

        function [] = ComputeExpectError(this)            
            
            J = numel(this.s_hat);            
            
            % sum_t rho_ij(t)
            sumRho = sum(this.rho,3); % I x J
            
            % term 1
            epsilon = this.x - squeeze(sum(this.sim,2)); % T x I

            % term 2
            normS = sum(this.s.^2,1); % 1 x J
            normS_sumRho = permute(bsxfun(@times, normS, sumRho), [2,1]); % J x I
            normS_sumRho = sum(normS_sumRho,1)';
            
            % term 3
            gamma_sumRho = 0;
            for j=1:J
                gamma_sumRho = gamma_sumRho + permute( sum(sum(bsxfun(@times, this.gamma{j}, permute(sumRho(:,j), [2,3,1]) ), 1), 2), [3,1,2]); % I x 1
            end

            % term 4
            normG = cellfun(@(x) sum(x.^2,2), this.G, 'UniformOutput', false); % {Jx1}[F x 1 x I]
            gamma_normG = 0;
            for j=1:J
                gamma_normG = gamma_normG + permute(sum(sum(bsxfun(@times, this.gamma{j}, normG{j}), 1), 2), [3,1,2]); % I x 1
            end
            
            this.expectError = sum(epsilon.^2,1).' + normS_sumRho + gamma_sumRho + gamma_normG;
            
        end
                
        function [likelihood_term, s_term, v_term, a_term, u_term] = ComputeVarFreeEnergy(this, b_computeExpectError)
            
            if nargin < 2
                b_computeExpectError = 1;
            end
            
            [T,I] = size(this.x);
            J = numel(this.s_hat);            
            F = cellfun('size', this.s_hat, 1);
            N = cellfun('size', this.s_hat, 2);
            La = size(this.a_hat,3);
            
            if b_computeExpectError
                this.ComputeExpectError;
            end

            likelihood_term = -I*T*log(2*pi)/2 - T/2*sum(log(this.sigma2)) - ...
                1/2*sum(this.expectError./this.sigma2);
            
            s_term = sum(F.*N)/2 + sum(F.*N)/2*psi(this.nu_v);
            for j=1:J
                s_term = s_term - 0.5*sum( log(this.beta{j}(:)./(this.gamma{j}(:) + eps) + eps) + log(this.lambda2{j}(:) + eps) + ...
                    this.nu_v./this.beta{j}(:).*(this.s_hat{j}(:).^2 + this.gamma{j}(:))./(this.lambda2{j}(:) + eps) );
            end
            
            v_term = -sum(F.*N)*( gammaln(this.alpha_v/2) + this.alpha_v/2*log(2/this.alpha_v) + ...
                psi(this.nu_v)*(this.nu_v - this.alpha_v/2) - this.nu_v - gammaln(this.nu_v) );
            for j=1:J
                v_term = v_term - this.alpha_v/2*sum( log(this.beta{j}(:)) + this.nu_v./this.beta{j}(:) );
            end
            
            a_term = I*J*La/2 - I*J/2*sum(log(this.q2) - psi(this.nu_u)) - ...
                0.5*sum(sum(sum( log(this.d./this.rho) + ...
                bsxfun(@times, (this.a_hat.^2 + this.rho)./this.d, permute(this.nu_u./this.q2, [2,3,1])) )));
            
            u_term = -I*J*sum( gammaln(this.alpha_u/2) + this.alpha_u/2.*log(2./this.alpha_u) + ...
                psi(this.nu_u).*(this.nu_u - this.alpha_u/2) - this.nu_u - gammaln(this.nu_u) ) - ...
                sum( this.alpha_u/2.* ( permute( sum( sum( log(this.d), 1), 2),  [3,1,2]) + ...
                this.nu_u.*permute( sum( sum( 1./this.d, 1), 2),  [3,1,2]) ) ); 
            
            this.varFreeEnergy = likelihood_term + s_term + v_term + a_term + u_term ;
            
        end

        function [] = ComputeGMat(this)
            
            J = numel(this.s_hat);            
            F = cellfun('size', this.s_hat, 1);
            [~,I] = size(this.x);
            
            for j=1:J
                win = sin((.5:this.wlen(j)-.5)/this.wlen(j)*pi); % sine window
                a_sq = permute(this.a_hat(:,j,:),[3,1,2]); % La x I
                for f=0:F(j)-1     
                    psi_f = win.*( sqrt(2/F(j)) * cos( (f+1/2) * ((0:this.wlen(j)-1) + 0.5 + this.wlen(j)/4) * 2*pi/this.wlen(j) ) ); % wlen x 1
                    this.G{j}(f+1,:,:) = this.FFTConv(psi_f'*ones(1,I),a_sq); % (La + wlen - 1) x I
                end
            end
        end        
        
        function [] = ImdctMultiProp(this, srcInd)

            J = numel(this.s_hat);            
            N = cellfun('size', this.s_hat, 2);

            Ls = size(this.s,1);
            
            if nargin<2
                srcInd = 1:J;
            end
            
            hop = this.wlen/2;

            for j = srcInd
                
                X_ext = [this.s_hat{j} ; -this.s_hat{j}(end:-1:1,:)];
                x_rec = zeros(hop(j)*(N(j)+1), 1);
                win = sin((.5:this.wlen(j)-.5)/this.wlen(j)*pi); % sine window
                
                frames = real(ifft(X_ext.*(exp(1i*2*pi*(0.5+this.wlen(j)/4)*(0:this.wlen(j)-1)'/this.wlen(j))*ones(1,N(j))))...
                    .*(exp(1i*pi*(0.5+this.wlen(j)/4+(0:this.wlen(j)-1)')/this.wlen(j))*ones(1,N(j))))*sqrt(this.wlen(j));
                for n = 1:N(j)
                    frameInd = (n-1)*hop(j) + (1:this.wlen(j));
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                this.s(:,j) = x_rec(1:Ls,:,:);
            end
        end
        
        function [x] = ImdctMulti(this, X)
            
            J = numel(X);            
            N = cellfun('size', X, 2);
            
            Ls = size(this.s,1);
            x = zeros(Ls,J);
            
            hop = this.wlen/2;
            
            for j = 1:J
                
                X_ext = [X{j} ; -X{j}(end:-1:1,:)];
                x_rec = zeros(hop(j)*(N(j)+1), 1);
                win = sin((.5:this.wlen(j)-.5)/this.wlen(j)*pi); % sine window
                
                frames = real(ifft(X_ext.*(exp(1i*2*pi*(0.5+this.wlen(j)/4)*(0:this.wlen(j)-1)'/this.wlen(j))*ones(1,N(j))))...
                    .*(exp(1i*pi*(0.5+this.wlen(j)/4+(0:this.wlen(j)-1)')/this.wlen(j))*ones(1,N(j))))*sqrt(this.wlen(j));
                for n = 1:N(j)
                    frameInd = (n-1)*hop(j) + (1:this.wlen(j));
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                x(:,j) = x_rec(1:Ls,:,:);
            end
        end
        
        function [] = SrcImageProp(this, srcInd, micInd)
            % s : [Ls x J]
            % si : [T x J x I]
            
            J = length(srcInd);
            I = length(micInd);
            T = size(this.x,1);
            Ls = size(this.s,1);
            La = size(this.a_hat,3);
            
            s_loc = this.s(:, srcInd);
            a = this.a_hat(micInd, srcInd, :);
            
            s_resh = zeros(Ls,J,I);
            for i=1:I
                s_resh(:,:,i) = s_loc;
            end
            s_resh = reshape(s_resh, Ls, I*J);
            a_resh = reshape(permute(a,[3,2,1]), La, I*J);
            si = this.FFTConv(s_resh,a_resh);
            this.sim(:,srcInd, micInd) = reshape(si, T, J, I);
            
        end
        
        function [xi] = SrcImage(this, x)
            % s : [Ls x J]
            % si : [T x J x I]
            
            J = size(x,2);
            [T, I] = size(this.x);
            Ls = size(this.s,1);
            La = size(this.a_hat,3);
                                    
            x_resh = zeros(Ls,J,I);
            for i=1:I
                x_resh(:,:,i) = x;
            end
            x_resh = reshape(x_resh, Ls, I*J);
            a_resh = reshape(permute(this.a_hat,[3,2,1]), La, I*J);
            xi = this.FFTConv(x_resh,a_resh);
            xi = reshape(xi, T, J, I);
            
        end

    end
    
    methods(Static)
        
        function v = cell2vec(c)
            v = cell2mat(cellfun(@(x) x(:), c, 'uniformoutput', false));
        end
        
        function [ psi ] = GetMDCTAtom(f, n, siglen, wlen )
            
            F = wlen/2;
            win = sin((.5:wlen-.5)/wlen*pi); % sine window
            hop = wlen/2; % hop size = wlen/2;
            
            frameInd = (n*hop+1 : n*hop + wlen)';
            if frameInd(end)>siglen
                error('out of signal')
            end
            psi = zeros(siglen,1);
            psi(frameInd) = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) );
            
        end
        
        function y = FFTConv(x, h)
            
            T = length(x) + length(h) - 1;
            Nfft = 2^nextpow2(T);
            y = real(ifft(fft(x, Nfft).*fft(h, Nfft), Nfft));
            y = y(1:T,:);
            
        end
        
    end
end

