function [ x ] = imdct_multi( X, siglen, b_fft, L_zp )

%IMDCT_MULTI Multichannel Inverse Modified Discrete Cosine Transform (IMDCT)
% using half-overlapping sine windows.
%
% x = istft_multi(X, siglen)
% x = istft_multi(X, siglen, b_fft)
%
% Inputs:
% X: [F x N x J] matrix containing the MDCT coefficients for J
% sources with F frequency bins and N time frames or [F x N x J x I]
% matrix containing the MDCT coefficients of J spatial
% source images over I channels.
% siglen: number of samples to which the corresponding time-domain signals
% are to be truncated.
% b_fft: if 1, fft is used to compute MDCT.
%
% Output:
% x: [siglen x J] matrix or [siglen x J x I] matrix containing the
% corresponding time-domain signals.
% If x is a set of signals of length siglen and X = mdct_multi(x), then
% x = imdct_multi(X,siglen).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simon Leglaive, Tlcom ParisTech, 2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin<2
    error('Not enough input arguments.');
end
if nargin<3
    b_fft = 0;
end

[F,N,J,I] = size(X);
wlen = 2*F; % window length

if nargin<4
    L_zp = wlen;
end

hop = wlen/2; % hop size = wlen/2;
win = sin((.5:wlen-.5)/wlen*pi); % sine window

x = zeros(hop*(N+1), J, I);

if ~b_fft
    
    % [F x wlen] matrix of MDCT transform for a frame
    T = zeros(F,wlen);
    for f = 0:F-1,
        T(f+1,:) = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) );
    end
    
    %%%%%%% LOOP %%%%%%%
    for i=1:I
        for j = 1:J
            frames = T'*X(:,:,j,i);
            for n = 1:N
                frameInd = (n-1)*hop + (1:wlen);
                x(frameInd,j,i) = x(frameInd,j,i) + frames(:,n);
            end
        end
    end
else
    
    X_ext = [X(:,:,:,:) ; -X(end:-1:1,:,:,:)];
    
    for i=1:I
        for j = 1:J
            x_ext = real(ifft(X_ext(:,:,j,i).*repmat(exp(1i*2*pi*(0.5+wlen/4)*(0:wlen-1)'/wlen),1,N))...
                .*repmat(exp(1i*pi*(0.5+wlen/4+(0:wlen-1)')/wlen),1,N))*sqrt(wlen);
            for n = 1:N
                frameInd = (n-1)*hop + (1:wlen);
                x(frameInd,j,i) = x(frameInd,j,i) + win'.*x_ext(:,n);
            end
        end
    end
    
end

x = x((L_zp/2+1):(L_zp/2+siglen),:,:);

end

