function [W, H, E] = t_nmf(V, W0, H0, phi, alpha, iter_num)

%
% [W, H, E] = t_nmf(V, W0, H0, iter_num);
%
% NMF decomposition using multiplicative update (MU) rules under a
% Student's t model
%
%
% input 
% -----
%
% V                 : power spectrogram [F x N]
% W0                : initial matrix of bases [F x K]
% H0                : initial matrix of activations [K x N]
% phi               : if phi = 2 (resp. 1) the underlying TF transform is
%                     an MDCT (resp. OFSTFT)
% alpha             : Student's t model shape parameter
% iter_num          : (opt) number of iterations (def = 100)
%
% output
% ------
%
% W                 : estimated matrix of bases [F x K]
% H                 : estimated matrix of contributions [K x N]
% E                 : vector of approximation errors (per iteration)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin < 6 || isempty(iter_num)
    iter_num = 100;
end;

[F,N] = size(V);

W = W0;
H = H0;

matOfOnes = ones(F,N);
WH = W*H + eps;

P = 1./( ( phi*alpha*(V + eps).^(-1) + 2*(WH + eps).^(-1) ) / ( phi*alpha + 2 ) );
S = sum(P(:));

E = zeros(1, iter_num);

for iter = 1:iter_num
    % update of W
    W = W .* (( P./((WH).^2) )*H' + eps) ./ ((1./(WH))*H' + eps);    
    
    % normalization
    w = sum(W,1); 
    d = diag(1./w); 
    W = W*d; % Normalizatiosn of columns of W
    d = diag(w); 
    H = d*H; % Energy transfer to H
    
    WH = W*H + eps;
    P = 1./( ( phi*alpha*(V + eps).^(-1) + 2*(WH + eps).^(-1) ) / ( phi*alpha + 2 ) );
    
    % update of H
    H = H .* (W'*(P./((WH).^2)) + eps) ./ (W'*(1./(WH)) + eps);   
    
    WH = W*H + eps;
    P = 1./( ( phi*alpha*(V + eps).^(-1) + 2*(WH + eps).^(-1) ) / ( phi*alpha + 2 ) );
    
    % Error with IS divergence:
    E(iter) = sum(sum(P./(WH + eps) - log(P./(WH + eps) + eps) - matOfOnes))/S; % normalized error
    
end
