%==========================================================================
% Main script running the VEM algorithm for multichannel audio source
% separation
%
% S. Leglaive, R. Badeau, G. Richard. "Student's t source and mixing models 
% for multichannel audio source separation", submitted for publication in 
% a journal, 2017. 
%
% Simon Leglaive, Telecom ParisTech, 2017
%==========================================================================

addpath('./aux_tools')
addpath('./mtimesx')
addpath('./toeplitzmult')

%% Global parameters

dataDir = './sounds'; % directory of the true source images and mix 
resultsDir = dataDir; % directory for writing estimated source images

st_mix.song = 'vieux_farka_toure-ana_201-213'; % name of the mixture
st_mix.sources = {'drums', 'vocals', 'bass'}; % sources in the mixture

fs = 16e3; % sampling rate
J = numel(st_mix.sources); % number of sources
I = 2; % number of microphones
niter = 200; % number of iterations for the VEM algorithm
T60 = 360; % reverberation time in ms
La = T60*1e-3*fs+1; % mixing filters length
wlen = zeros(J,1); % MDCT window length
isDrums = cellfun(@(x) strcmp(x, 'drums'), st_mix.sources);
wlen(isDrums) = 512;
wlen(~isDrums) = 2048;
hop = wlen/2;
varFreeEnergy = zeros(niter,1); % variational free energy
b_noiseAnnealing = 1; % indicates if noise annealing is used

%% Read mixture wav file

wavFile = fullfile(dataDir, 'mix.wav');
x = audioread(wavFile); 
T_orig = size(x,1); 
Ls = T_orig-La+1; % length of the source signals

%% Model parameters

% Noise annealing strategy
if b_noiseAnnealing
    sigma2_beg = 1e-2;
    sigma2_end = 1e-6;
    sigma2_init = sigma2_beg*ones(I,1);
else
    sigma2_init = sum(abs(x).^2,1)'/T_orig;
end

% Mixing filters model
tau = T60*1e-3*fs/(3*log(10)); % exponential decay factor
t_vec = (0:La-1)';
sigma_q = .1;
q2 = sigma_q^2*exp(-2*t_vec/tau); % Squared Student's t scale parameters [La x 1]
alpha_u = ones(La,1); % Student's t shape parameter [La x 1]

% Source model
K = 10; % NMF rank
alpha_v = 100; % Student's t shape parameter
W_init = cell(J,1); % spectral templates
H_init = cell(J,1); % temporal activations
for j=1:J
    wavFile = fullfile(dataDir,['src' num2str(j) '.wav']);
    [s, fs] = audioread(wavFile);
    s = s(1:Ls);
    S = mdct_multi(s,wlen(j),0, max(wlen));
    rng(1);
    W0 = 0.5 * (abs(randn(size(S,1),K)) + ones(size(S,1),K)) .* ...
        (mean(abs(S).^2, 2)  * ones(1,K));
    rng(1);
    H0 = 0.5 * (abs(randn(K,size(S,2))) + ones(K,size(S,2)));
    [W_init{j}, ~, ~] = t_nmf(abs(S).^2, W0, H0, 2, alpha_v, 200);
    H_init{j} = ones(size(H0));
end

%% Zero padding for handling edges

Nzp_end = ceil(Ls/max(hop))*max(hop)-Ls;
Nzp_end = Nzp_end + max(wlen)/2;
Nzp_beg = max(wlen)/2;
Ls = Ls + Nzp_beg + Nzp_end;
x = [zeros(Nzp_beg, I) ; x ; zeros(Nzp_end, I)];
T = size(x,1);
F = wlen/2; % number of time frames
N = round(Ls./hop) - 1; % number of frames

%% VEM algorithm

fprintf('Instantiation of VEMalgo\n');

ss_algoVEM = VEMalgo('x', x, 'alpha_u', alpha_u, 'q2', q2, 'sigma2', sigma2_init, ...
        'alpha_v', alpha_v, 'W', {W_init}, 'H', {H_init}, 'wlen', wlen, 'La', La);

for iter = 1:niter
    
    fprintf('\nVEM iteration %d of %d\n\n', iter, niter);
    
    fprintf('E-V Step\n');
    ss_algoVEM.E_V_Step;
    
    fprintf('E-S Step\n');
    ss_algoVEM.E_S_Step;
    
    fprintf('E-U Step\n');
    ss_algoVEM.E_U_Step;
    
    fprintf('E-A Step\n');
    ss_algoVEM.E_A_Step;
    
    ss_algoVEM.ComputeVarFreeEnergy(1);
    varFreeEnergy(iter) = ss_algoVEM.varFreeEnergy;
    
    fprintf('variational free energy : %.2f\n', varFreeEnergy(iter)/(I*T));
    
    fprintf('NMF Update\n');
    ss_algoVEM.UpdateLambda(0);
    
    fprintf('Noise Update\n');
    if b_noiseAnnealing
        ss_algoVEM.UpdateNoiseVarAnneal(iter, niter, sigma2_beg, sigma2_end);        
    else
        ss_algoVEM.ComputeExpectError;
        ss_algoVEM.UpdateNoiseVar;
    end
 
end

ss_algoVEM.E_V_Step;
ss_algoVEM.E_S_Step;
ss_algoVEM.E_U_Step;
ss_algoVEM.E_A_Step;

%% Compute estimated source images

Ls = T_orig-La+1;
se = zeros(Ls, J);
sime = zeros(J, T_orig, I);

for j=1:J
    
    se(:,j) = imdct_multi(ss_algoVEM.s_hat{j}, Ls, 0, max(wlen));
    
    for i=1:I
        sime(j,:,i) = conv(se(:,j), squeeze(ss_algoVEM.a_hat(i,j,:)));
    end
    
    audiowrite(fullfile(resultsDir, ['sim_VEM_' num2str(j) '.wav']), reshape(sime(j,:,:),T_orig,2),fs);
end

return

%% Compute scores with BSS eval

fprintf('Computation of the scores\n\n');

sim_true = zeros(J,T_orig,2);

for j=1:J
    sim_true(j,:,:) = reshape(audioread(fullfile(dataDir, ['sim' int2str(j) '.wav'])),1,T_orig,2);
end

[SDR_VEM, ISR_VEM, SIR_VEM, SAR_VEM] = bss_eval_images_nosort(sime, sim_true);